/*
 * RepositoryContainer
 * Copyright (C) 2011 by Apaq
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package dk.apaq.framework.repository;


import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import dk.apaq.framework.criteria.Criteria;
import dk.apaq.framework.criteria.Limit;
import java.util.Arrays;

/**
 * An abstract implementation of a <code>Repository</code> instance taking care
 * of RepositoryNotifer logic. An implementing class needs only to call the fire* methods(fx. fireOnRead)
 * and this class will make sure to fire evetns on the listeners.
 */
public abstract class BaseRepository<BT, IDT> implements
        Repository<BT, IDT>, RepositoryNotifier {

    private final List<WeakReference<RepositoryListener>> listeners = new ArrayList<WeakReference<RepositoryListener>>();

    /**
     * {@inheritDoc}
     */
    @Override
    public void addListener(RepositoryListener listener) {
        WeakReference<RepositoryListener> refListener = new WeakReference<RepositoryListener>(listener);
        listeners.add(refListener);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void removeListener(RepositoryListener listener) {
        synchronized(listeners) {
            Iterator<WeakReference<RepositoryListener>> it = listeners.iterator();
            while(it.hasNext()) {
                WeakReference<RepositoryListener> refListener = it.next();
                if(refListener.get() == listener) {
                    it.remove();
                    break;
                }
            }
        }
    }
    
    protected List<RepositoryListener> getListeners() {
        List<RepositoryListener> list = new ArrayList<RepositoryListener>();
        Iterator<WeakReference<RepositoryListener>> it = listeners.iterator();
        while(it.hasNext()) {
            WeakReference<RepositoryListener> refListener = it.next();
            if(refListener.get() == null) {
                it.remove();
               
            } else {
                list.add(refListener.get());
            }
        }
        return Collections.unmodifiableList(list);
    }

    
    /**
     * Tells listeners that an entity has been read from crud.
     * @param entity The entity that was read.
     */
    protected void fireOnRead(IDT id, BT entity) {
        RepositoryEvent.WithIdAndEntity e = createEventWithIdAndEntity(id, entity);
        for (RepositoryListener listener : getListeners()) {
            listener.onEntityRead(e);
        }
    }

    /**
     * Tells listeners that an entity is about tobe read.
     * @param id The id of the entity that is to be read.
     */
    protected void fireOnBeforeRead(IDT id) {
        RepositoryEvent.WithId e = createEventWithId(id);
        for (RepositoryListener listener : getListeners()) {
            listener.onBeforeEntityRead(e);
        }
    }
    
    /**
     * Tells listeners that a new entity is about to be created.
     */
    protected void fireOnBeforeCreate(BT entity) {
        RepositoryEvent.WithEntity e = createEventWithEntity(entity);

        for (RepositoryListener listener : getListeners()) {
            listener.onBeforeEntityCreate(e);
        }
    }

    /**
     * Tells listeners that an entity is about to be updated.
     * @param entity The entity that holds changes and is about to be persisted.
     */
    protected void fireOnBeforeUpdate(IDT id, BT entity) {
        RepositoryEvent.WithIdAndEntity e = createEventWithIdAndEntity(id, entity);

        for (RepositoryListener listener : getListeners()) {
            listener.onBeforeEntityUpdate(e);
        }
    }

    /**
     * Tells listeners that an eneity is about to be deleted.
     * @param id The id of the entity that is about to be deleted.
     */
    protected void fireOnBeforeDelete(IDT id) {
        RepositoryEvent.WithId e = createEventWithId(id);
        for (RepositoryListener listener : getListeners()) {
            listener.onBeforeEntityDelete(e);
        }
    }

    /**
     * Tells listeners that an entity has been created.
     * @param entity The entity that was created.
     */
    protected void fireOnCreate(IDT id, BT entity) {
        RepositoryEvent.WithIdAndEntity e = createEventWithIdAndEntity(id, entity);

        for (RepositoryListener listener : getListeners()) {
            listener.onEntityCreate(e);
        }
    }

    /**
     * Tells listeners that an entity was updated.
     * @param entity The entity that was updated.
     */
    protected void fireOnUpdate(IDT id, BT entity) {
        RepositoryEvent.WithIdAndEntity e = createEventWithIdAndEntity(id, entity);

        for (RepositoryListener listener : getListeners()) {
            listener.onEntityUpdate(e);
        }
    }

    /**
     * Tells listeners that an entity was deleted.
     * @param id The id of the entity that was deleted.
     */
    protected void fireOnDelete(IDT id) {
        RepositoryEvent.WithId e = createEventWithId(id);
        for (RepositoryListener listener : getListeners()) {
            listener.onEntityDelete(e);
        }
    }

    /**
     * Tells listeners that an entity is about tobe read.
     * @param id The id of the entity that is to be read.
     */
    protected void fireOnBeforeList(Criteria criteria) {
        RepositoryEvent.List e = new RepositoryEvent.List<BT, IDT>(this, criteria);
        for (RepositoryListener listener : getListeners()) {
            listener.onBeforeList(e);
        }
    }
    
    public BT findFirst(Criteria criteria) {
        criteria = criteria == null ? new Criteria() : criteria;
        Limit limit = criteria.getLimit() == null ? new Limit(0, 1) : new Limit(criteria.getLimit().getOffset(), 1);
        criteria = new Criteria(criteria.getRule(), criteria.getSorter(), limit);
        
        List<BT> result = findAll(criteria);
        return result.isEmpty() ? null : result.get(0);
    }

    public List<BT> findAll(IDT... ids) {
        return findAll(Arrays.asList(ids));
    }

    public List<BT> findAll(List<IDT> ids) {
        List<BT> beans = new ArrayList();
        for(IDT id : ids) {
            beans.add(findOne(id));
        }
        return beans;
    }

    
    public List<BT> findAll() {
        return findAll((Criteria)null);
    }
    
    public List<IDT> findAllIds() {
        return findAllIds(null);
    }
    
    public void deleteAll(List<BT> entities) {
        for(BT entity : entities) {
            delete(entity);
        }
    }

    public void deleteAll(BT... entities) {
        for(BT entity : entities) {
            delete(entity);
        }
    }
    
    public void deleteAllById(List<IDT> ids) {
        for(IDT id:ids) {
            deleteById(id);
        }
    }

    public void deleteAllById(IDT... ids) {
        for(IDT id:ids) {
            deleteById(id);
        }
    }
    
    protected RepositoryEvent createEvent() {
        return new RepositoryEvent(this);
    }

    protected RepositoryEvent.WithId<BT, IDT> createEventWithId(IDT id) {
        return new RepositoryEvent.WithId<BT, IDT>(this, id);
    }

    protected RepositoryEvent.WithEntity<BT, IDT> createEventWithEntity(BT entity) {
        return new RepositoryEvent.WithEntity<BT, IDT>(this,entity);
    }
    
    protected RepositoryEvent.WithIdAndEntity<BT, IDT> createEventWithIdAndEntity(IDT id, BT entity) {
        return new RepositoryEvent.WithIdAndEntity<BT, IDT>(this, id, entity);
    }

}
